#ifndef _INCLUDE_CODE_CONTEXT_H_
#define _INCLUDE_CODE_CONTEXT_H_


#include <cassert>
#include <strstream>
#include "LocalVars.h"

class CfNeedy;
class CtRegistry;
class File;
class TreeNode;

extern ostream &endCline( ostream & );


class CodeContext : public ostrstream, public LocalVars {

public:
  explicit CodeContext( CodeContext & );
  explicit CodeContext( CodeContext &, ostream & );
  explicit CodeContext( CfCode & );
  ~CodeContext();
  
  void setDebugLine(TreeNode *t);
  int isValidDebug() const {
    return ((curr_debug_line != -1) &&
	    (curr_debug_file != NULL));
  }
  int getCurrDebugLine() const {
    assert(isValidDebug());
    return curr_debug_line;
  }
  File *getCurrDebugFile() const {
    assert(isValidDebug());
    return curr_debug_file;
  }
  int getStartDebugLine() const {
    assert(isValidDebug());
    return start_debug_line;
  }
  File *getStartDebugFile() const {
    assert(isValidDebug());
    return start_debug_file;
  }

  static CodeContext *recover( const ostream & );

private:
  ostream &sink;
  CodeContext *parent;
  TreeNode *lastNodeSeen;

  const int level;
  int curr_debug_line, start_debug_line;
  File *curr_debug_file, *start_debug_file;

  void codeGen() const;
  void codeGen( ostream & ) const;

  static const int recoverSlot;

  // forbidden
  bool operator ==( const CodeContext & ) const;
  static void *operator new( size_t );
};


#endif // !_INCLUDE_CODE_CONTEXT_H_
