#include <fstream.h>
#include "AST.h"
#include "CodeContext.h"
#include "CfHeader.h"
#include "CfSource.h"
#include "aux-code.h"
#include "decls.h"
#include "threading.h"


extern bool opt_dataflow;


void InterfaceDeclNode::codeGen()
{
  ClassDecl &declaration = *decl();
  const CtType &descriptorType = declaration.cDescriptorType();
  const string  descriptorName = declaration.cDescriptorName();
  const string className = decl()->cType().name;
  
  // Declarations
  {
    CfHeader os( declaration.cType() );
    
    declaration.cType().define( os );
    os << "extern " << descriptorType << ' ' << descriptorName << ";\n";
    
    os << "\nextern void " << className << "_class_init();\n";
  }

  // Definitions
  {
    CfSource os( declaration.cType() );
    
    const char *listName = declaration.emitInterfaceList( os );
    os << descriptorType << ' ' << descriptorName << " = {\n"
       << "  Interface, " << listName << "\n};\n\n";
    
    emitStaticFields(declaration, staticFieldDecl, os);
    os << "\nvoid " << className << "_class_init()\n";
    {
      CodeContext subcontext( os );
      vector< TreeNode * > inits;

      foriter (member, members()->allChildren(), ChildIter)
	if ((*member)->decl()->modifiers() & Static)
	  inits.push_back( *member );
	
      if (opt_dataflow)
	EncapsulateInits(inits)->emitStatement(subcontext);
      else
	codeGenAll(subcontext, inits);
    }
  }
}
