#include "AST.h"
#include "CodeContext.h"
#include "CfSource.h"
#include "CtType.h"
#include "MethodDecl.h"
#include "code.h"
#include "code-assign.h"
#include "decls.h"


void FieldDeclNode::codeGen( CodeContext &context )
{
  TypeNode *itype;
  string initializer; 
  if (initExpr()->absent()) {
    if (dtype()->isTitaniumArrayType() || dtype()->isImmutable()) {
      extern const string callNoArgConstructor(TypeNode *t);

      assert(dtype()->isTitaniumArrayType() || dtype()->isImmutable());
      initializer = callNoArgConstructor(dtype());
    } else if (dtype()->isReference())
      initExpr(new CastNode(new NullPntrNode(), dtype()));
    else if (dtype()->isPrimitive())
      initExpr(new CastNode(new LitNode(intLiteral(string("0"))), dtype()));
    else
      assert(0);
    itype = dtype();
  } else
    itype = initExpr()->type();

  if (initializer == "")
    initializer = initExpr()->emitExpression( context );

  if ((decl()->modifiers() & Static))
    context << makeAssign("&STATIC_REF(" + cFieldName(decl()) + ")",
			  makeCast(itype, initializer),
			  true, assignKind(itype), true) << ";" << endCline;
  else
    context << "me->" << cFieldName(decl()) << " = " <<
      makeCast(itype, initializer) << ";" << endCline;
#if 0
    if (decl()->container()->asType()->isImmutable())
      context << "me." << cFieldName(decl()) << " = "
	      << makeCast(itype, initializer)
		<< ";" << endCline;
    else
      context << makeAssign("&me->" + cFieldName(decl()),
			    makeCast(itype, initializer),
			    true, assignKind(itype), false)
			      << ";" << endCline;
#endif
}

extern bool opt_dataflow;

void VarDeclNode::codeGen( CodeContext &context )
{
  // We cannot just initialize directly due to C++ restrictions on
  // skipping over initializations when jumping to "case" labels.
  // If you don't believe me, see CD2,  6.7 [stmt.dcl],  3.

  const string varName = "var_" + *simpName()->simpName()->ident();
  context.declare( varName, dtype()->cType() );

  string initial;
  bool have_initial = true;

  if (initExpr()->absent()) {
    // from code-alloc.cc
    extern const string callNoArgConstructor(TypeNode *t);

    if (dtype()->isTitaniumArrayType() || dtype()->isImmutable()) {
      initial = callNoArgConstructor(dtype());
    }
    else {
      have_initial = false;
    }
  } 
  else {
    initial = initExpr()->emitExpression( context );
  }

  if (have_initial) {
    context << varName << " = " << initial << ";" << endCline;
  }
  else {
    assert(opt_dataflow);
    context << varName << ";" << endCline;
  }
}


void MethodDeclNode::codeGen( CfSource &os )
{
  MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
  
  if (!(mdecl.modifiers() & (Common::Native | Common::Abstract))) {
    bool first = true;

    os << returnType()->cType() << ' '
       << mdecl.cMethodNameStatic() << '(';
    
    const TypeNode *thisType = mdecl.thisType();
    if (thisType) {
      os << thisType->cType() << " me";
      first = false;
    }	

    foriter (param, params()->allChildren(), TreeNode::ChildIter) {
      if (first) first = false; else os << ",";
      os << (*param)->dtype()->cType()
	 << " var_" << *(*param)->simpName()->ident();
    }

    os << ")\n";
    {
      CodeContext context( os );
      body()->codeGen( context );
    }
    os << '\n';
  }
}


void MethodSignatureNode::codeGen ( CodeContext & )
{
}


void ConstructorDeclNode::codeGen( CfSource &os )
{
  const MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
  const TypeNode &thisType = *mdecl.thisType();
  const string thisTypeName = thisType.cType().name;
  
  os << thisTypeName << ' '
     << decl()->cMethodNameStatic()
     << "(";

  bool first = true;
  if (!thisType.isImmutable()) {
    os << thisTypeName << " me";
    first = false;
  }

  foriter (param, params()->allChildren(), TreeNode::ChildIter) {
    if (!first) os << ",";
    os << (*param)->dtype()->cType()
       << " var_" << *(*param)->simpName()->ident();
    first = false;
  }
  os << ")\n";

  {
    CodeContext context( os );
    const MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
    const TypeNode &thisType = *mdecl.thisType();
    
    if (thisType.isImmutable()) {
      // declare me
      context.declare( "me", thisType.cType() );
      // zero out the contents of me
      context << "memset( &me, 0, sizeof(me) );" << endCline
	      << decl()->container()->cInitFieldsName()
	      << "( &me );" << endCline;
    } else {
      constructorCall()->codeGen( context );
      context << endCline;
    }
    
    body()->codeGen( context );
    context << "return me;" << endCline;
  }
  os << "\n\n";
}
