(defun defattr* (name type value)
  (setq *attributes* (cons (list name type value) *attributes*))
  (insert (format "%s TreeNode::%s () const\n" (to-c++-node-type type) name)
	  "{\n"
	  (format "    undefined (\"%s\");\n" name)
	  (format "    return %s;\n" value)
	  "}\n"
	  (format "void TreeNode::%s (%s)\n" name (to-c++-node-type type))
	  (format "{ undefined (\"%s\"); }\n" name)))

(defun defattr-default* (name type value)
  (setq *attributes* (cons (list name type) *attributes*))
  (insert (format "%s TreeNode::%s () const\n" type name)
	  "{\n"
	  (format "    return %s;\n" value)
	  "}\n"
	  (format "void TreeNode::%s (%s)\n" name type)
	  (format "{ undefined (\"%s\"); }\n" name)))

(defun defchild* (name type)
  (insert (format "%s TreeNode::%s () const\n" (to-c++-node-type type) name)
	  (format "{ undefined(\"%s\"); return 0; }\n" name)
	  (format "void TreeNode::%s (%s)\n" name (to-c++-node-type type))
	  (format "{ undefined(\"%s\"); }\n" name)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(unless (fboundp 'string-prefix-p)
  ;; 24 Oct 1997 gnu.emacs.help
  ;; Philippe Schnoebelen <phs@mirabelle.lsv.ens-cachan.fr>
  (defun string-prefix-p (s1 s2)
    "True if string S1 is a prefix of S2 (i.e. S2 starts with S1)"
    (equal 0 (string-match (regexp-quote s1) s2))))

(defun initialized-states (states)
  (and states
       (let ((state (car states))
	     (rest (initialized-states (cdr states))))
	 (if (caddr state)
	     (cons state rest)
	   rest))))

(defun defnode* (name parents fields methods states)
  (let* ((details (resolve-fields fields))
	 (flds (aref details 2))
	 (kids (aref details 0))
	 (attrs (aref details 1))
	 (states (resolve-states states))
	 (init-states (initialized-states states)))
    (insert
     (format "\n\n\t/* Class %s */\n" name)
     (format "int %s::arity () const { return %d; }\n\n"
	     name (length kids))
     (format "const char* %s::oper_name () const { return \"%s\"; }\n\n"
	     name name)
     (format "%s *%s::clone () const { %s* copy = new %s(*this); copy->children = %s; return copy; }\n"
	     name name name name (if kids "copy->_children" "NULL"))
     (format "%s::%s(" name name))
    (loop for f in flds do
	  (insert (format "%s __%s, " (to-c++-type (cadr f)) (car f))))
    (insert "SourcePosn _posn)\n")
    (insert-initializers attrs init-states t)
    (insert "{\n")
    (if kids 
	(insert "    children = _children;\n")
        (insert "    children = NULL;\n"))
    (loop for i = 0 then (1+ i)
	  for k in kids do
	  (if (list-type (cadr k))
	      (insert 
	       (format "    child(%d, new %s(__%s));\n"
		       i (cadr k) (car k)))
	    (insert (format "    child(%d, __%s);\n" i (car k)))))
    (insert "    setWhere (_posn);\n")
    (insert "}\n\n")
    (when init-states
      (insert (format "%s::%s ()\n" name name))
      (insert-initializers attrs init-states nil)
      (insert "{\n}\n\n"))
    (if attrs
	(create-print-method name flds))
    (create-compare-method name flds states)
    ))
	 
;;; type is the C type.  var is the name of the variable or accessor method,
;;; e.g., "imports()"
(defun create-comparison (nodetype type var)
  (insert (format "  // nodetype=%s, type=%s, var=%s\n" nodetype type var))
  (cond ((string-match "TreeListNode" nodetype)
	 (insert (format "COMPARE(%s, u->%s)" var var)))
	((and (string-match "TreeNode *\\*" type)
	      (string-match "list" type))
	 (insert (format "compareLists(%s, u->%s)" var var)))
	((string-match "TreeNode *\\*" type)
	 (insert (format "(%s->absent() == u->%s->absent()) && (%s->absent() || COMPARE(%s, u->%s))" var var var var var)))
	((and (string-match "TypeNode *\\*" type)
	      (not (string-match "list" type)))
	 (insert (format "(%s->cType().name == u->%s->cType().name)" var var)))
	((string-match "Literal" type)
	 (insert (format "(%s.asString() == u->%s.asString())" var var)))
	((string-match "string *\\*" type)
	 (insert (format "(*%s == *u->%s)" var var)))
	(t
	 (insert (format "(%s == u->%s)" var var)))))

(defun create-compare-method (name flds states)
  (insert
   (format "    bool %s::compare (const TreeNode *t) const\n" name)
   "{\n"
   "  if (!streq(oper_name(), t->oper_name())) return false;\n"
   (format "  const %s *u = (const %s *) t;\n" name name)
   "  return true ")
  (loop for f in flds do
	(insert (format "\n     && // fld=%s\n" f))
	(create-comparison (to-c++-node-type (cadr f))
			   (to-c++-type (cadr f)) (format "%s()" (car f))))
  (loop for info in states do
	(unless (string-prefix-p "static " (cadr info))
	  (insert (format "\n     && // state=%s\n" info))
	  (create-comparison (cadr info) (cadr info) (car info))))
  (insert ";\n}\n"))

(defun create-print-method (name flds)
  (insert
   (format "void %s::print (ostream& os, unsigned depth) const\n" name)
   "{\n"
   (format "    indent(os, depth);\n")
   (format "    os << \"(%s\";\n" name)
   (format "    os << '<';\n")
   (format "    if(position().file) os << *position().file->name(); else os << \"unknown\";\n")
   (format "    os << ' ' << position().posn << ',' << (position().posn+1) << '>';\n"))
  (loop for f in flds do
	(cond ((equal (cadr f) "const string*")
	       (insert (format "    os << \" \\\"\" << *%s() << \"\\\"\";\n" 
			       (car f))))
	      ((or (equal (cadr f) "Common::Modifiers")
		   (equal (cadr f) "Modifiers"))
	       (insert "    os << \" (\";\n"
		       (format "    printModifiers (os, %s());\n" (car f))
		       "    os << \")\";\n"))
	      ((or (list-type (cadr f))
		   (and (symbolp (cadr f)) (string-match "Node" (symbol-name (cadr f)))))
	       (if (eq (car f) 'destination)
		   (insert "    os << \"-*-\";\n")
		 (insert "    os << \"\\n\";\n"
			 (format "    %s()->print (os, depth + 1);\n" (car f)))))
	      ((and (stringp (cadr f)) (string-match "Decl\*" (cadr f)))
	       (insert "    os << \" \";\n"
		       (format "    %s()->print (os);\n" (car f))))
	      (t 
	       (insert (format "    os << \" \" << %s();\n" (car f))))))
  (insert "    os << \")\";\n"
	  "}\n")
  )

(defun insert-initializers (attrs states withargs)
  (and (or attrs states)
       (progn (insert "    : ")
	      (let ((sep ""))
		(loop for f in attrs do
		      (if (list-type (cadr f))
			  (insert (format "%s_%s(new %s(__%s))\n"
					  sep (car f) (cadr f) (car f)))
			(insert
			 (if withargs
			     (format "%s_%s(__%s)" sep (car f) (car f))
			   (format "%s_%s(%s)" sep (car f) (caddr f)))))
		      (setq sep ",\n"))
		(loop for f in states do
		      (insert (format "%s%s(%s)\n" sep (car f) (caddr f)))
		      (setq sep ","))))))

(defun deflistnode* (name methods)
  (insert 
   (format "\n\n\t/* Class %s */\n" name)
   (format "%s *%s::clone () const\n{\n" name name)
   (format "  %s* copy = new %s; *copy = *this;\n" name name)
   "  copy->children = new TreeNode*[_arity];\n"
   "  memcpy(copy->children, children, _arity * sizeof(*children));\n"
   "  return copy;\n}\n\n"))
