#include <cctype>
#include <cstdlib>
#include <iomanip.h>
#include <strstream.h>
#include "errors.h"
#include "lex-string.h"
#include "string16.h"


static uint16 parseInteger( const char *&str, const unsigned base, const unsigned maxDigits )
{
    unsigned long value;
    istrstream source( str, maxDigits );
    const streampos begin = source.tellg();

    source >> setbase( base ) >> value;

    const streampos end = source.tellg();
    str += end - begin;

    return value;
}


static uint16 convert( char *&str, const SourcePosn &position )
{
    switch (*str) {
    case '\\':
	switch (*++str) {
	case 'n': ++str; return '\n';
	case 'b': ++str; return '\b';
	case 'r': ++str; return '\r';
	case 't': ++str; return '\t';
	case 'f': ++str; return '\f';
	case '\\': case '"': case '\'':
	    return *str++;
	case '0': case '1': case '2':
	    return parseInteger( str, 8, 3 );
	case '3': case '4': case '5': case '6': case '7':
	    return parseInteger( str, 8, 2 );
	case 'u': {
	    do
		str++;
	    while (*str == 'u');
	    const char * const before = str;
	    const uint16 result = parseInteger( str, 16, 4 );
	    // require exactly four hex digits
	    if (str - before == 4)
		return result;
	}
	default:
	    Error("illegal escape sequence in literal", position);
	    return '?';
	}
    default:
	return *str++;
    }
}


uint16 convertCharacter( char *str, size_t length, const SourcePosn &position )
{
    str[ length ] = '\0';
    return convert( str, position );
}


const string16 convertString( char* str, size_t length, const SourcePosn &position )
{
    str[ length ] = '\0';

    string16 result;
    result.reserve( length );

    const char * const end = str + length;
    while (str < end)
	result += convert( str, position );

    return result;
}
