#include "AST.h"
#include "decls.h"
#include "solver.h"


void AssignNode::inferLocal( const LocalContext )
{
  opnd0()->localReceiveAssign( *opnd1()->type() );
  type()->localIdenticalTo( *opnd0()->type() );
}


////////////////////////////////////////////////////////////////////////


void TreeNode::localReceiveAssign( const TypeNode & ) const
{
  undefined( "localReceiveAssign" );
}


////////////////////////////////////////////////////////////////////////


void ArrayAccessNode::localReceiveAssign( const TypeNode &donor ) const
{
  const TypeNode &outer = *array()->type();
  const TypeNode &inner = *outer.elementType();

  donor.localWeaklySubsumedBy( inner );

  if (outer.localInferable() && inner.localInferable())
    solver().subsumedBy( outer, inner );
}


void FieldAccessNode::localReceiveAssign( const TypeNode &donor ) const
{
  donor.localWeaklySubsumedBy( *decl()->type() );
}


void ObjectFieldAccessNode::localReceiveAssign( const TypeNode &donor ) const
{
  FieldAccessNode::localReceiveAssign( donor );

  // !! need to go farther out in the case of "object.immutable.field"
  const TypeNode &outer = *accessedObjectType();
  const TypeNode &inner = *decl()->type();
  if (outer.localInferable() && inner.localInferable())
    solver().subsumedBy( outer, inner );
}


void ObjectNode::localReceiveAssign( const TypeNode &donor ) const
{
  donor.localWeaklySubsumedBy( *decl()->type() );
}


void PointArrayAccessNode::localReceiveAssign( const TypeNode & ) const
{
}
