#include "AST.h"
#include "decls.h"
#include "solver.h"


void TreeListNode::localConstrainCall( const TypeListNode &formal ) const
{
  assert( arity() == formal.arity() );
  
  for (unsigned sweep = formal.arity(); sweep--; )
    child( sweep )->type()->localWeaklySubsumedBy( *formal.child( sweep ) );
}


void TreeNode::localConstrainCall( const TypeNode &callee ) const
{
  args()->localConstrainCall( *callee.paramTypes() );
}


////////////////////////////////////////////////////////////////////////


void MethodCallAssignNode::inferLocal( const LocalContext )
{
  const TypeNode &dispatcher = *method()->accessedObjectType();
  if (dispatcher.localInferable())
    solver().subsumedBy( dispatcher, *method()->decl()->type() );
    
  const TypeNode &target = *method()->type();
  localConstrainCall( target );
  type()->localIdenticalTo( *target.returnType() );
}


void MethodCallNode::inferLocal( const LocalContext )
{
  const TypeNode &dispatcher = *method()->accessedObjectType();
  if (dispatcher.localInferable())
    solver().subsumedBy( dispatcher, *method()->decl()->type() );
    
  const TypeNode &target = *method()->type();
  localConstrainCall( target );
  type()->localIdenticalTo( *target.returnType() );
}


void SuperConstructorCallNode::inferLocal( const LocalContext )
{
  localConstrainCall( *decl()->type() );
}


void ThisConstructorCallNode::inferLocal( const LocalContext )
{
  localConstrainCall( *decl()->type() );
}
