#include "AST.h"
#include "decls.h"
#include "solver.h"


void ExprNode::inferLocal( const LocalContext )
{
  assert( !type()->localInferable() );
}


////////////////////////////////////////////////////////////////////////


void AllocateArrayNode::inferLocal( const LocalContext )
{
  const TypeNode *stack = type();

  // !! Is this order correct, or is it backwards?
  for (unsigned sweep = dimExprs()->arity(); sweep--; )
    {
      solver().identicalTo( *stack, *dimExprs()->child( sweep ) );
      stack = stack->elementType();
    }

  dtype()->localIdenticalTo( *stack );
}


void AllocateNode::inferLocal( const LocalContext )
{
  localConstrainCall( *decl()->type() );
}


void ArrayInitNode::inferLocal( const LocalContext )
{
  const TypeNode &elemType = *type()->elementType();
  
  foriter (initializer, initializers()->allChildren(), ChildIter)
    (*initializer)->type()->localWeaklySubsumedBy( elemType );
}


void BroadcastNode::inferLocal( const LocalContext )
{
  expr()->type()->localWeaklySubsumedBy( *type() );
  
  if (type()->localInferable())
    solver().mustBeGlobal( *type() );
}


void EqualityNode::inferLocal( const LocalContext )
{
  opnd0()->type()->localSimilarTo( *opnd1()->type() );
}


void IfExprNode::inferLocal( const LocalContext )
{
  thenOpnd()->type()->localWeaklySubsumedBy( *type() );
  elseOpnd()->type()->localWeaklySubsumedBy( *type() );
}


void NullPntrNode::inferLocal( const LocalContext )
{
}


void ObjectNode::inferLocal( const LocalContext )
{
  type()->localIdenticalTo( *decl()->type() );
}


void StringConcatAssignNode::inferLocal( const LocalContext )
{
  type()->localIdenticalTo( *opnd0()->type() );
}


void StringConcatNode::inferLocal( const LocalContext )
{
}


void StringLitNode::inferLocal( const LocalContext )
{
}


void ThisNode::inferLocal( const LocalContext context )
{
  if (context && type()->localInferable())
    solver().identicalTo( *context, *type() );
}
