#include "AST.h"
#include "context.h"
#include "enact-context.h"
#include "options.h"
#include "solver.h"
#include "stats.h"
#include "typecheck.h"
#include "utils.h"


void inferLocal()
{
  if (opt_local) {
    llist< TreeNode * > * const allInsts = templateEnv.allInstances();
    llist< TreeNode * > * const allTrees = extend( allInsts, allFiles );

    LocalSolver solver;
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->unshareTypeTree();

    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->inferLocalTree();

    const LocalSolution solution( solver.solution() );
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->localEnactTree( solution );

    // This is a minor hack.  Ordinarily, an initial type context is
    // created by CompileUnitNode::typecheck().  However, the
    // instantiated template classes from "allInsts" are not contained
    // within CompileUnitNodes.  So we need to provide a dummy initial
    // type context in which these classes may be checked.
    
    TreeNode::TypeContext typeContext;
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->typecheck( &typeContext );

    LocalEnactContext::printStats( cerr );
  }
}
