#include <cstdio>
#include <iostream.h>
#include "AST.h"
#include "code-foreach.h"
#include "code.h"
#include "config.h"
#include "decls.h"
#include "errors.h"
#include "local/options.h"
#include "utils.h"

int DEBUG = 0;
bool compilerLocComments = true;
bool debug_opt = false;
bool debug_commonsub;
bool debug_defs = false;
bool opt_osr = true;
bool opt_foreach = true;
bool opt_lift = true;
bool opt_dataflow = false;
bool test_sr = false;
bool cheating = false;
bool codeGen_libs = true;
bool codeGen_main = true;
bool use_sglobal_inference = true;
bool use_goto = false;
bool use_nonTrivialSizeCheck = false;
bool force_java = false, force_ti = false;
string codeGen_outdir = "./";
int debugger_level = 0;
bool line_directives = true;
string dump_ast;

main(int argc, char* argv[])
{
  bool frontend_only = false;

  while (argc > 1 && argv[1][0] == '-') {
    if (strcmp(argv[1], "-fonly") == 0) {
      frontend_only = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-g") == 0) {
      debugger_level = 1;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-noline") == 0) {
      line_directives = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debug") == 0) {
      DEBUG = 1;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-trivialSizeCheck") == 0) {
      use_nonTrivialSizeCheck = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-notrivialSizeCheck") == 0) {
      use_nonTrivialSizeCheck = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-goto") == 0) {
      use_goto = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-nogoto") == 0) {
      use_goto = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debugopt") == 0) {
      debug_opt = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debugdefs") == 0) {
      debug_defs = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-nocheat") == 0) {
      cheating = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-cheat") == 0) {
      cheating = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-srtest") == 0) {
      test_sr = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-stats") == 0) {
      stat_foreach = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-O0") == 0) {
      opt_osr = false;
      opt_foreach = false;
      opt_lift = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-O") == 0 || strcmp(argv[1], "-O1") == 0) {
      opt_foreach = true;
      opt_lift = true;
      opt_osr = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Oforeach") == 0) {
      opt_foreach = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onoforeach") == 0) {
      opt_foreach = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OOSR") == 0) {
      opt_osr = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OnoOSR") == 0) {
      opt_osr = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Olift") == 0) {
      opt_lift = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onolift") == 0) {
      opt_lift = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Odataflow") == 0) {
      opt_dataflow = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onodataflow") == 0) {
      opt_dataflow = false;
      argc--;
      argv++;
    }
#ifdef USE_LOCAL_INFERENCE
    else if (strcmp(argv[1], "-Olocal") == 0) {
      opt_local = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onolocal") == 0) {
      opt_local = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-ibane") == 0) {
      opt_ibane = argv[2];
      argc -= 2;
      argv += 2;
    }
#endif // USE_LOCAL_INFERENCE
#ifdef USE_SGLOBAL_INFERENCE
    else if (strcmp(argv[1], "-infer-sglobal") == 0) {
	use_sglobal_inference = true;
	argc--;
	argv++;
    }
#endif // USE_SGLOBAL_INFERENCE
    else if (strcmp(argv[1], "-noinfer-sglobal") == 0) {
	use_sglobal_inference = false;
	argc--;
	argv++;
    }
    else if (strcmp(argv[1], "-nolibs") == 0) {
      codeGen_libs = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-libs") == 0) {
      codeGen_libs = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-nomain") == 0) {
      codeGen_main = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-main") == 0) {
      codeGen_main = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-outdir") == 0) {
      codeGen_outdir = argv[2];
      codeGen_outdir += '/';
      argc -= 2;
      argv += 2;
    }
    else if (strcmp(argv[1], "-dumpast") == 0) {
      dump_ast = argv[2];
      argc -= 2;
      argv += 2;
    }
    else if (strcmp(argv[1], "-java") == 0) {
      force_java = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-ti") == 0) {
      force_ti = true;
      argc--;
      argv++;
    }
    else {
      cerr << "Ignoring unknown option: " << argv[1] << '\n';
      argc--;
      argv++;
    }      
  }

#ifdef USE_LOCAL_INFERENCE
  // Local inference does not yet cope with separate compilation.
  if (opt_local && !codeGen_libs)
    {
      cerr <<
	"Local qualification inference does not yet handle separate compilation.\n"
	"Please use tcbuild option \"--rebuild-tlib\" or tc option \"-libs\"."
	   << endl;
      exit(1);
    }
#endif // USE_LOCAL_INFERENCE

  // If testing strength reduction, make sure we do that optimization.
  if (test_sr) opt_foreach = true;

  debug_commonsub = false; 
  // debug_commonsub = debug_opt;

  initTheSystem();

  if (argc > 1)
    for (int source = 1; source < argc; ++source)
      load(argv[source], !force_java);
           //!force_java && (force_ti || isTitanium(argv[source])));
  else
    load("<stdin>", stdin, !force_java);

  staticSemantics();

  if (!frontend_only && NumErrors() == 0) {
    if (opt_foreach) foreachOptPass();
    codeGen();
  }

  showStats();
  exit(NumErrors() > 0 ? 1 : 0);
}
