#include "file-64.h"		/* must appear before fcntl & unistd */

#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "filio.h"
#include "layout!GP_jbyte.h"
#include "layout!GP_jint.h"
#include "native-file-utils.h"
#include "native-utils.h"
#include "runtime-options.h"

typedef GP_cl_java_io_FileDescriptor GP_FileDesc;
typedef GP_type(GP_FileDesc) GP_GP_FileDesc;


static GP_FileDesc getFileDescriptor( GP_cl_java_io_FileInputStream me )
{
    GP_GP_FileDesc field;
    GP_FileDesc fileDescriptor;

    CHECK_NULL_GLOBAL( me );

    FIELD_ADDR_GLOBAL( field, me, cl_java_io_FileInputStream_fld_fd );
    DEREF_GLOBAL( fileDescriptor, field );
    return fileDescriptor;
}


static int getFd( GP_cl_java_io_FileInputStream me )
{
    GP_FileDesc fileDescriptor;
    GP_jint field;
    jint fd;

    CHECK_NULL_GLOBAL( me );
    
    fileDescriptor = getFileDescriptor( me );
    CHECK_NULL_GLOBAL( fileDescriptor );
    
    FIELD_ADDR_GLOBAL( field, fileDescriptor, cl_java_io_FileDescriptor_fld_fd );
    DEREF_GLOBAL( fd, field );
    return fd;
}


jint cl_java_io_FileInputStream_mt_available_G( GP_cl_java_io_FileInputStream me )
{
    const int fd = getFd( me );
    
    int available;
    
    if (ioctl( fd, FIONREAD, &available ))
	tossIOException();
    else
	return available;
}


void cl_java_io_FileInputStream_mt_close_G( GP_cl_java_io_FileInputStream me )
{
    CHECK_NULL_GLOBAL( me );
    jclose( getFileDescriptor( me ) );
}


void cl_java_io_FileInputStream_mt_open_G_GP_cl_java_lang_String(
    GP_cl_java_io_FileInputStream me,
    GP_cl_java_lang_String name )
{
    CHECK_NULL_GLOBAL( me );
    jopen( getFileDescriptor( me ), name, O_RDONLY );
}


jint cl_java_io_FileInputStream_mt_read_G( GP_cl_java_io_FileInputStream me )
{
    const int fd = getFd( me );
    
    jbyte byte;
    switch (read( fd, &byte, 1 )) {
    case 0:
	return -1;
    case 1:
	return byte;
    default:
	tossIOException();
    }
}


jint cl_java_io_FileInputStream_mt_readBytes_G_GP_JA_jbyte_jint_jint(
    GP_cl_java_io_FileInputStream me,
    GP_JA_jbyte b, jint off, jint len )
{
    const int fd = getFd( me );
    JAVA_ARRAY_CHECK_GLOBAL( b, off, len );
    
    {
	VARARRAY_DECLARE(jbyte, buffer, len);
	const jint received = read( fd, buffer, len );

	VARARRAY_CREATE(jbyte, buffer, len);
	if (received < 0) {
	    VARARRAY_DESTROY(buffer);
	    tossIOException();
	} else {
	    GP_jbyte rem_buf;
	    JAVA_ARRAY_ADDR_GLOBAL( rem_buf, b, off );
	    /* rem_buf.processor( b.processor() ); */
	    /* rem_buf.address( b.address()->data + off ); */
	    local_to_global_copy(buffer, rem_buf, len);
	    VARARRAY_DESTROY(buffer);
	    return received;
	}
	VARARRAY_DESTROY(buffer);
    }
}


jlong cl_java_io_FileInputStream_mt_skip_G_jlong(
    GP_cl_java_io_FileInputStream me, jlong n )
{
    const int fd = getFd( me );
    
    if (lseek64( fd, n, SEEK_CUR ) == -1)
	tossIOException();
    else
	return n;
}
