#include <fcntl.h>
#include <unistd.h>
#include "layout!GP_jbyte.h"
#include "layout!GP_jint.h"
#include "native-file-utils.h"
#include "native-utils.h"
#include "runtime-options.h"

typedef GP_cl_java_io_FileDescriptor GP_FileDesc;
typedef GP_type(GP_FileDesc) GP_GP_FileDesc;


static GP_cl_java_io_FileDescriptor getFileDescriptor( GP_cl_java_io_FileOutputStream me )
{
    GP_GP_FileDesc field;
    GP_FileDesc fileDescriptor;

    CHECK_NULL_GLOBAL( me );

    FIELD_ADDR_GLOBAL( field, me, cl_java_io_FileOutputStream_fld_fd );
    DEREF_GLOBAL( fileDescriptor, field );
    return fileDescriptor;
}


static int getFd( GP_cl_java_io_FileOutputStream me )
{
    GP_FileDesc fileDescriptor;
    GP_jint field;
    jint fd;

    CHECK_NULL_GLOBAL( me );
    
    fileDescriptor = getFileDescriptor( me );
    CHECK_NULL_GLOBAL( fileDescriptor );
    
    FIELD_ADDR_GLOBAL( field, fileDescriptor, cl_java_io_FileDescriptor_fld_fd );
    DEREF_GLOBAL( fd, field );
    return fd;
}


void cl_java_io_FileOutputStream_mt_close_G( GP_cl_java_io_FileOutputStream me )
{
    CHECK_NULL_GLOBAL( me );
    jclose( getFileDescriptor( me ) );
}


void cl_java_io_FileOutputStream_mt_open_G_GP_cl_java_lang_String(
    GP_cl_java_io_FileOutputStream me,
    GP_cl_java_lang_String name )
{
    CHECK_NULL_GLOBAL( me );    
    jopen( getFileDescriptor( me ), name,
	   O_CREAT | O_WRONLY | O_TRUNC );
}


void cl_java_io_FileOutputStream_mt_writeBytes_G_GP_JA_jbyte_jint_jint(
    GP_cl_java_io_FileOutputStream me,
    GP_JA_jbyte b, jint off, jint len)
{
    const int fd = getFd( me );
    JAVA_ARRAY_CHECK_GLOBAL( b, off, len );

    {
        VARARRAY_DECLARE(jbyte, buffer, len);
	GP_jbyte rem_buf;
	const jbyte *current = buffer;

        VARARRAY_CREATE(jbyte, buffer, len);
	JAVA_ARRAY_ADDR_GLOBAL(rem_buf, b, off); /* global pointer to remote buffer */
	/* rem_buf.processor = ( b.processor ); */
	/* rem_buf.address = ( b.address->data + off ); */

	global_to_local_copy(rem_buf, buffer, len);

	while (len > 0) {
	    const ssize_t chunk = write( fd, current, len );
	    if (chunk >= 0) {
		len -= chunk;
		current += chunk;
	    } else {
                VARARRAY_DESTROY(buffer);
		tossIOException();
	    }
	}
        VARARRAY_DESTROY(buffer);
    }
}


void cl_java_io_FileOutputStream_mt_write_G_jint(
    GP_cl_java_io_FileOutputStream me, jint byte )
{
    const int fd = getFd( me );
    const char buffer = byte;

    /* We expect that write() of one byte will never return 0. */
    if (write( fd, &buffer, 1 ) <= 0)
	tossIOException();
}
