#include <stdlib.h>
#include <string.h>

#include <class_header.h>
#include <native-stubs.h>
#include <ti-memory.h>
#include "cl_java_lang_Cloneable.h"

#ifdef HAVE_MONITORS
#include "layout!GP_LP_titanium_monitor_t.h"
#endif


typedef GP_cl_java_lang_Object GP_Object;
typedef LP_cl_java_lang_Object LP_Object;

jboolean cl_java_lang_Object_mt_isLocal_G(GP_cl_java_lang_Object me) {
  return isLocal(me);
}

jint cl_java_lang_Object_mt_creator_G(GP_cl_java_lang_Object me) {
  CHECK_NULL_GLOBAL(me);
  return TO_PROC(me);
}

jint cl_java_lang_Object_mt_hashCode_G (GP_cl_java_lang_Object p) {
  return TO_PROC(p) ^ ((int) TO_LOCAL(p));
}

void cl_java_lang_Object_mt_object_initialize_L(LP_cl_java_lang_Object p) {
  p->monitor = monitor_init();
}

void cl_java_lang_Object_mt_object_finalize_G(GP_cl_java_lang_Object p) {
#ifdef HAVE_MONITORS
  GP_LP_titanium_monitor_t fieldAddr;
  LP_titanium_monitor_t monitor;
  
  FIELD_ADDR_GLOBAL(fieldAddr, p, monitor);
  DEREF_GLOBAL(monitor, fieldAddr);
  monitor_destroy(monitor);
#endif /* monitors */
}

void cl_java_lang_Object_mt_notify_G(GP_cl_java_lang_Object p) {
#ifdef HAVE_MONITORS
  GP_LP_titanium_monitor_t fieldAddr;
  LP_titanium_monitor_t monitor;
  
  FIELD_ADDR_GLOBAL(fieldAddr, p, monitor);
  DEREF_GLOBAL(monitor, fieldAddr);
  monitorNotify(monitor, TO_PROC(p));
#else /* no monitors */
  fputs( "unimplemented native method: cl_java_lang_Object_mt_notify_G\n", stderr );
  exit(2);
#endif /* no monitors */
}

void cl_java_lang_Object_mt_notifyAll_G(GP_cl_java_lang_Object p) {
#ifdef HAVE_MONITORS
  GP_LP_titanium_monitor_t fieldAddr;
  LP_titanium_monitor_t monitor;
  
  FIELD_ADDR_GLOBAL(fieldAddr, p, monitor);
  DEREF_GLOBAL(monitor, fieldAddr);
  monitorNotifyAll(monitor, TO_PROC(p));
#else /* no monitors */
  fputs( "unimplemented native method: cl_java_lang_Object_mt_notifyAll_G\n", stderr );
  exit(2);
#endif /* no monitors */
}

void cl_java_lang_Object_mt_wait_G_jlong(GP_cl_java_lang_Object p, jlong millis) {
#ifdef HAVE_MONITORS
  GP_LP_titanium_monitor_t fieldAddr;
  LP_titanium_monitor_t monitor;
  
  FIELD_ADDR_GLOBAL(fieldAddr, p, monitor);
  DEREF_GLOBAL(monitor, fieldAddr);
  monitorWait(monitor, millis, TO_PROC(p));
#else /* no monitors */
  fputs( "unimplemented native method: cl_java_lang_Object_mt_wait_G_jlong\n", stderr );
  exit(2);
#endif /* no monitors */
}


Region cl_java_lang_Object_mt_regionOf_G(GP_Object me) {
  return PG2Region(me);
}


static size_t allocateClone( const cl_java_lang_Object_desc *info, LP_Object *duplicate )
{
  if (CLASS_IMPLEMENTS( info, cl_java_lang_Cloneable_descriptor ))
    {
      *duplicate = (LP_Object) malloc( info->size );
      return info->size;
    }
  else
    {
      cl_ti_lang_NativeUtils_mt_throwCloneNotSupportedException();
      return 0;
    }
}


LP_Object cl_java_lang_Object_mt_localClone_L( LP_Object me )
{
  const cl_java_lang_Object_desc * const info = me->class_info;
  LP_Object duplicate;
  const size_t size = allocateClone( info, &duplicate );
  
  memcpy( duplicate, me, size );
  return duplicate;
}


#ifdef ALLOW_CLONE_EVEN_THOUGH_ITS_SEMANTICS_MAY_CHANGE
LP_Object cl_java_lang_Object_mt_clone_G( GP_Object me )
{
  cl_java_lang_Object_desc *info;
  CLASS_INFO_GLOBAL( info, cl_java_lang_Object_desc, me );
  
  {
    LP_Object duplicate;
    const size_t size = allocateClone( info, &duplicate );
    char * const bytes = (char *) duplicate;
  
    global_to_local_copy( me, bytes, size );
    info->nullifyLocalFields( duplicate );
    return duplicate;
  }
}
#else
unimplemented(cl_java_lang_Object_mt_clone_G, LP_Object, (GP_Object me))
#endif

     
unimplemented(cl_java_lang_Object_mt_getClass_G, GP_cl_java_lang_Class, (GP_Object me))
