#ifndef _INCLUDE_TLIB_PRIMITIVES_H
#define _INCLUDE_TLIB_PRIMITIVES_H

#include "config.h"

/* Define the ABI for use by the Titanium runtime. This should match
   the ABI specified by Java.

   Item         Description
   ----         -----------
   jboolean     An unsigned boolean
   jchar        An unsignedcharacter
   jbyte        A signed byte
   jshort       A short signed integer
   jint         A signed integer
   jlong        A long signed integer

   jIntPointer  A signed integer that holds a pointer
   jUIntPointer An unsigned integer that holds a pointer

   jfloat       A single precision floating point number
   jdouble      A double precision floating point number

   */

/* ------------------
   Special Platforms
   ------------------ */

/* These platforms are handled specially. If a special platform is handled
   here, it should declare all of the types specified above.  Also, it
   should set the define:

   #define _INCLUDE_TLIB_PRIMITIVES_H_SPECIAL_PLATFORM

   To ensure that the Standard platform portion of this file is skipped.
   */

/* The Tera */
#if defined(__TERA__)
typedef unsigned      char jboolean;
typedef unsigned __short16 jchar;
typedef   signed      char jbyte;
typedef          __short16 jshort;
typedef            __int32 jint;
typedef            __int64 jlong;
typedef            __int64 jIntPointer;
typedef unsigned   __int64 jUIntPointer;
typedef              float jfloat;
typedef             double jdouble;
#  define _INCLUDE_TLIB_PRIMITIVES_H_SPECIAL_PLATFORM_SET
#endif /* defined(__TERA__) */

/* --------------------
   Compiler Information
   -------------------- */

/* Provide some information about the compiler. Here are some examples
   that need to be set for certain compilers.
   
   Use "long long" as a 64-bit number. This must be defined if the LLP64,
   ILP32, or LP32 ABI's are used.
   #define _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64 long long

   Use "__int32" as a 32-bit number in the ILP64 ABI.
   #define _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT32 __int32

   Force the use of the ILP32 ABI assuming there was not match otherwise.
   #define _INCLUDE_TLIB_PRIMITIVES_H_ILP32

   */

/* Check for any ABI related defines preset by the compiler */
#if defined(_LP64) || defined(__LP64) || defined(__LP64__)
#  define _INCLUDE_TLIB_PRIMITIVES_H_LP64
#  define _INCLUDE_TLIB_PRIMITIVES_H_ABI_SET
#endif
#if defined(_ILP64) || defined(__ILP64) || defined(__ILP64__)
#  define _INCLUDE_TLIB_PRIMITIVES_H_ILP64
#  define _INCLUDE_TLIB_PRIMITIVES_H_ABI_SET
#endif
#if defined(_LLP64) || defined(__LLP64) || defined(__LLP64__)
#  define _INCLUDE_TLIB_PRIMITIVES_H_LLP64
#  define _INCLUDE_TLIB_PRIMITIVES_H_ABI_SET
#endif
#if defined(_ILP32) || defined(__ILP32) || defined(__ILP32__)
#  define _INCLUDE_TLIB_PRIMITIVES_H_ILP32
#  define _INCLUDE_TLIB_PRIMITIVES_H_ABI_SET
#endif
#if defined(_LP32) || defined(__LP32) || defined(__LP32__)
#  define _INCLUDE_TLIB_PRIMITIVES_H_LP32
#  define _INCLUDE_TLIB_PRIMITIVES_H_ABI_SET
#endif

/* For now, assume long long, a ILP32 ABI, and __int32 is the default.
   The configure script should really do a better job of finding the right
   answers. */
#if !defined(_INCLUDE_TLIB_PRIMITIVES_H_ABI_SET)
#  define _INCLUDE_TLIB_PRIMITIVES_H_ILP32
#endif
#if !defined(_INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64)
#  define _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64 long long
#endif
#if !defined(_INCLUDE_TLIB_PRIMITIVES_H_NONSTD_UINT64)
#  define _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_UINT64 unsigned long long
#endif
#if !defined(_INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT32)
#  define _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT32 __int32
#endif

/* ------------------
   Standard Platforms
   ------------------ */

/* The integers and pointers */

#if !defined(_INCLUDE_TLIB_PRIMITIVES_H_SPECIAL_PLATFORM_SET)

#  if defined(USE_RUNTIME_INTTYPES_H)

/* The standard inttypes.h */
#    include <inttypes.h>
typedef        uint8_t jboolean;
typedef       uint16_t jchar;
typedef         int8_t jbyte;
typedef        int16_t jshort;
typedef        int32_t jint;
typedef        int64_t jlong;
typedef       intptr_t jIntPointer;
typedef      uintptr_t jUIntPointer;

#  else /* defined(USE_RUNTIME_INTTYPES_H) */
#    if defined(_INCLUDE_TLIB_PRIMITIVES_H_LP64)

/* LP64 - long, pointer: 64; int: 32; short: 16; char: 8 */
typedef unsigned  char jboolean;
typedef unsigned short jchar;
typedef   signed  char jbyte;
typedef          short jshort;
typedef            int jint;
typedef           long jlong;
typedef           long jIntPointer;
typedef unsigned  long jUIntPointer;

#    else /* LP64 */
#      if defined(_INCLUDE_TLIB_PRIMITIVES_H_ILP64)

/* ILP64 - int, long, pointer: 64; short: 16; char: 8 */
typedef unsigned  char jboolean;
typedef unsigned short jchar;
typedef   signed  char jbyte;
typedef          short jshort;
typedef _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT32 jint;
typedef           long jlong;
typedef           long jIntPointer;
typedef unsigned  long jUIntPointer;

#      else /* ILP64 */
#        if defined(_INCLUDE_TLIB_PRIMITIVES_H_LLP64)

/* LLP64 - long long, pointer: 64; int, long: 32; short: 16; char: 8 */
typedef     unsigned  char jboolean;
typedef     unsigned short jchar;
typedef       signed  char jbyte;
typedef              short jshort;
typedef                int jint;
typedef _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64 jlong;
typedef _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64 jIntPointer;
typedef _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_UINT64 jUIntPointer;

#        else /* LLP64 */
#          if defined(_INCLUDE_TLIB_PRIMITIVES_H_ILP32)

/* ILP32 - long long/<none>: 64; int, long, pointer: 32; short: 16; char: 8 */
typedef unsigned  char jboolean;
typedef unsigned short jchar;
typedef   signed  char jbyte;
typedef          short jshort;
typedef            int jint;
typedef _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64 jlong;
typedef            int jIntPointer;
typedef unsigned   int jUIntPointer;

#          else /* ILP32 */
#            if defined(_INCLUDE_TLIB_PRIMITIVES_H_LP32)

/* LP32 - <none>: 64; long, pointer: 32; int, short: 16; char: 8 */
typedef unsigned  char jboolean;
typedef unsigned short jchar;
typedef   signed  char jbyte;
typedef          short jshort;
typedef           long jint;
typedef _INCLUDE_TLIB_PRIMITIVES_H_NONSTD_INT64 jlong;
typedef           long jIntPointer;
typedef unsigned  long jUIntPointer;

#            else /* LP32 */
#              error PRIMITIVES_H Error: Unsupported ABI
#            endif /* LP32 */
#          endif /* ILP32 */
#        endif /* LLP64 */
#      endif /* ILP64 */
#    endif /* LP64 */
#  endif /* defined(USE_RUNTIME_INTTYPES_H) */

/* The floating point numbers */

typedef          float jfloat;
typedef         double jdouble;

#endif /* !defined(_INCLUDE_TLIB_PRIMITIVES_H_SPECIAL_PLATFORM_SET) */

#endif /* _INCLUDE_TLIB_PRIMITIVES_H */
