/* ti_array.c */
/* This file assumes T is a type and N is an arity. */

/* To add a method you must (as of 12/97):
     edit code-grid.cc, adding a line to the part that looks like
        D("DOMAIN_METHOD", "_ti_arraymethod_domain(" << t << ", " << n << ")",
     edit tlib/ti/internal/tiArray.java    
     edit titanium.h
     add the method defn to this file or ti_array.h
     */

#include "runtime-options.h"

#include <gc.h>
#include <primitives.h>
#include <regions.h>
#include <ti-memory.h>

#if (N == 2) || (N == 3)
#include <slice.h>
#endif

#define N_LARGER_THAN_1 N_MINUS_1
#define N_IS_1 (N && !N_LARGER_THAN_1)

#ifndef GLOBAL_ARRAY
#define GLOBAL_ARRAY 1
#endif

#ifndef BOUNDS_CHECKING
#define BOUNDS_CHECKING 1
#endif

#ifndef UNINITIALIZED_CHECKING
#define UNINITIALIZED_CHECKING BOUNDS_CHECKING
#endif

#ifndef DEBUG_ARRAYS
#define DEBUG_ARRAYS 0
#endif

#ifndef ASSERT_SAME_BYTES
#define ASSERT_SAME_BYTES
static void assertSameBytes(char *x, char *y, int len)
{
  while (len-- > 0) assert(*x++ == *y++);
}
#endif /* ASSERT_SAME_BYTES */

#ifndef DUMPABLE_ARRAYS
#define DUMPABLE_ARRAYS 0
#endif

#ifndef DEFAULT_DUMP_DEST
#define DEFAULT_DUMP_DEST "|/disks/orodruin/titanium/work/pike/bin/aview"
#endif

#if GLOBAL_ARRAY
#define deref DEREF_GLOBAL
#define INDEX INDEX_GLOBAL
#define LOCAL_PART TO_LOCAL
#define PROC_PART  TO_PROC
#else
#define deref DEREF_LOCAL
#define INDEX INDEX_LOCAL
#define LOCAL_PART
#define PROC_PART(p) MYPROC
#endif

#if GLOBAL_ARRAY && !GLOBAL_ELEMENTS
# define assignable 0
#else
# define assignable 1
#endif

#if ELEMENTS_ARE_ATOMIC
# define array_ralloc ralloc_atomic
# define array_malloc ti_malloc_atomic_huge
#else
# define array_ralloc ralloc
# define array_malloc ti_malloc_huge
#endif

#undef assertExists
#if UNINITIALIZED_CHECKING
#define assertExists(x, where)					\
  if ((x).stride[0] == 0) {					\
    fprintf(stderr, "fatal error: "				\
	    "attempt to use uninitialized array at %s\n",	\
	    where);						\
    assert(0);							\
    exit(-1);							\
  }
#else
#define assertExists(x, where)
#endif

#define PRINT_POINT(p, f) 					\
do {								\
  jint _i;							\
  fprintf(f, "[");						\
  for (_i = 1; _i <= N; _i++)					\
    fprintf(f, (_i > 1) ? ", %d" : "%d", POINT_INDEX(p, _i));	\
  fprintf(f, "]");						\
} while (0)

#define PRINT_DOMAIN(R, f)			\
do {						\
  jint _i;					\
  ti_POINT p = RECTDOMAIN_STRIDE(R);		\
  fputs("[", f);				\
  PRINT_POINT(RECTDOMAIN_MIN(R), f);		\
  fputs(" : ", f);				\
  PRINT_POINT(RECTDOMAIN_UPB(R), f);		\
  for (_i = 1; _i <= N; _i++)			\
    if (POINT_INDEX(p, _i) >= 2) break;		\
  if (_i <= N) {				\
    fputs(" : ", f);				\
    PRINT_POINT(p, f);				\
  }						\
  fputs("]", f);				\
} while (0)

#if DUMPABLE_ARRAYS
void ti_dump(ti_ARRAY *px, char *dest, char *userArg)
{
  ti_ARRAY x = *px;
  FILE *f;
  ti_RECTDOMAIN R = x.domain;
  ti_POINT p, q, s;
  jint lo[N], hi[N], stride[N], i;

  if (dest == NULL || dest[0] == '\0') dest = DEFAULT_DUMP_DEST;
  
  if (dest[0] == '|')
    f = popen(dest + 1, "w");
  else
    f = fopen(dest, "w");

  if (f == NULL) {
    fprintf(stderr, "Unable to open `%s'; dump request failed.\n", dest);
    return;
  }

  if (userArg != NULL) fprintf(f, "%s\n", userArg);
  fprintf(f, "%d\n", N);
  PRINT_DOMAIN(x.domain, f);
  fprintf(f, "\ndomainaddr <%d, %x>", TO_PROC(x.domain), TO_LOCAL(x.domain));
  fprintf(f, "\nstride");
  for (i = 0; i < N; i++) fprintf(f, " %d", x.stride[i]);
  fprintf(f, "\nbase");
  for (i = 0; i < N; i++) fprintf(f, " %d", x.base[i]);
  fprintf(f, "\nsideFactors");
  for (i = 0; i < N; i++) fprintf(f, " %d", x.sideFactors[i]);
  fprintf(f, "\ndesc %lx\nA %lx\nancestor %lx\nwhere %s\n",
	  (long)px, (long)LOCAL_PART(x.A), (long)LOCAL_PART(x.ancestor),
	  x.where);
  fprintf(f, "elementSize %d\n", sizeof(T));
  fputs("elementType " T_string "\n", f);

  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  
  forall(e, x, lo, hi, stride) {
    fputc('!', f);
    PRINT_FORALL_POINT(f);
    fputc(',', f);
    if (strcmp("jint", T_string) == 0)
      fprintf(f, "%d", deref(e));
    else if (strcmp("jdouble", T_string) == 0)
      fprintf(f, "%g", deref(e));
    else if (strcmp("jfloat", T_string) == 0)
      fprintf(f, "%f", deref(e));
    else {
      T buf = deref(e);
      for (i = 0; i < sizeof(T); i++)
	fprintf(f, "%02x", (unsigned int)(((unsigned char *) &buf)[i]));
    }
    fputc('!', f);
  }

  if (dest[0] == '|')
    pclose(f);
  else
    fclose(f);
}
#endif

jboolean ti_isnull(ti_ARRAY x)
{
  return (LOCAL_PART(x.A) == 0);
}

ti_ARRAY ti_empty()
{
  ti_ARRAY x;
  memset(&x, 0, sizeof(x));
  return x;
}

ti_ARRAY ti_construct(const char *where, Region pregion, ti_RECTDOMAIN R)
{
  ti_ARRAY r;
  jint i, max[N], size;
  ti_POINT p;
  T *array;

  r.domain = R;
  r.where = where;
#if DUMPABLE_ARRAYS
  r.dumpfn = (void(*)()) ti_dump;
#endif
  
  p = RECTDOMAIN_MIN(R); /* minimum point in R */
  for (i = 0; i < N; i++)
    r.base[i] = POINT_INDEX(p, i + 1);
  
  p = RECTDOMAIN_STRIDE(R); /* stride of R */
  for (i = 0; i < N; i++) 
    if ((r.stride[i] = POINT_INDEX(p, i + 1)) == 0) r.stride[i] = 1;
  
  p = RECTDOMAIN_MAX(R); /* maximum point in R */
  for (i = 0; i < N; i++)
    max[i] = (POINT_INDEX(p, i + 1) - r.base[i]) / r.stride[i];
    
  r.sideFactors[N - 1] = 1;
  if (N > 1) {
    jint sf = 1;
    for (i = N - 2; i >= 0; i--)
      r.sideFactors[i] = sf *= (1 + max[i + 1]);
  }

  size = 1 + max[0];
  if (N > 1)
    size *= r.sideFactors[0];

#if DEBUG_ARRAYS
  fprintf(stderr, "p%d: new array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(R), (int) TO_LOCAL(R));
  PRINT_DOMAIN(R, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tmax\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%dt%d\\n", i,
	    r.base[i], max[i], r.stride[i], r.sideFactors[i]);
  }
#endif

  if (pregion)
    array = (T *)array_ralloc(pregion, size * sizeof(T));
  else
    array = (T *)array_malloc(size * sizeof(T));

  memset(array, 0, size * sizeof(T));

#if GLOBAL_ARRAY
  TO_GLOBAL(r.A, MYPROC, array);
#else
  r.A = array;
#endif
  r.ancestor = r.A;

#if EXPLICITLY_STORE_CREATOR
  r.creator = MYPROC;
#endif

  return r;
}

#if BOUNDS_CHECKING
jint ti_conv(const char *where, const ti_ARRAY *x, ti_POINT p)
{
  jint sum = 0, q, qm, s;
#if N_LARGER_THAN_1
  jint i;
#else
#define i 1
#endif

  assertExists(*x, where);
#if N_LARGER_THAN_1
  for (i = N; i >= 1; i--)
#endif
    {
      q = POINT_INDEX(p, i);
      qm = q - x->base[i-1];
      if ((s = POINT_INDEX(RECTDOMAIN_STRIDE(x->domain), i)) > 1
	  && ((qm % s) != 0)) {
	fprintf(stderr, "bounds checking error at %s:\n  failed condition: [",
		where);
	PRINT_POINT(RECTDOMAIN_MIN(x->domain), stderr);
	fprintf(stderr, " : ");
	PRINT_POINT(RECTDOMAIN_UPB(x->domain), stderr);
	fprintf(stderr, " : ");
	PRINT_POINT(RECTDOMAIN_STRIDE(x->domain), stderr);
	fprintf(stderr, "].contains(");
	PRINT_POINT(p, stderr);
	fprintf(stderr, ")\n  array was allocated at %s\n", x->where);
	assert(("array index out of bounds", 0));
	exit(-2);
      }
      else if (q < POINT_INDEX(RECTDOMAIN_MIN(x->domain), i) ||
	       q > POINT_INDEX(RECTDOMAIN_MAX(x->domain), i)) {
	fprintf(stderr, "bounds checking error at %s:\n  failed condition: ",
		where);
	PRINT_POINT(RECTDOMAIN_MIN(x->domain), stderr);
	fprintf(stderr, " <= ");
	PRINT_POINT(p, stderr);
	fprintf(stderr, " <= ");
	PRINT_POINT(RECTDOMAIN_MAX(x->domain), stderr);
	fprintf(stderr, "\n  array was allocated at %s\n", x->where);
	assert(("array index out of bounds", 0));
	exit(-2);
      }
      /* At this point, x->stride[i-1] should divide qm.  Of course, we only
	 know for sure if bounds checking is on. */
      sum += (qm / x->stride[i-1]) * x->sideFactors[i-1];
    }
#if !N_LARGER_THAN_1
#undef i  
#endif

#if 0
  fprintf(stderr, "conv(");
  PRINT_POINT(p, stderr);
  fprintf(stderr, ") = %d\n", sum);
#endif
  return sum;
}
#endif /* BOUNDS_CHECKING */


#if N_LARGER_THAN_1
ti_ARRAY_SLICE SLICE_METHOD(const ti_ARRAY x, jint k, jint j)
{
  jint i, l;
  ti_ARRAY_SLICE r;
  ti_POINT p;

  assertExists(x, "(in slice method)");
  r.domain = RECTDOMAIN_SLICE(x.domain, k);
  r.where = x.where;

  for (i = l = 0; i < N_MINUS_1; i++, l++) {
    if (i == k - 1) ++l;
    r.stride[i] = x.stride[l];
    r.base[i] = x.base[l];
    r.sideFactors[i] = x.sideFactors[l];
  }    

#if BOUNDS_CHECKING
  p = POINT_SET(RECTDOMAIN_MIN(x.domain), k - 1, j);
  ti_conv("(in slice method)", &x, p); /* just for bounds check */
#endif
  INDEX(r.A, x.A, (j - x.base[k - 1]) / x.stride[k - 1] * x.sideFactors[k - 1]);
  
  r.ancestor = x.ancestor;
#if EXPLICITLY_STORE_CREATOR 
  r.creator = x.creator;
#endif

#if DEBUG_ARRAYS
  fprintf(stderr, "slice(%d, %d) of [%x]\n  Array<%d>:\tbase\tstride\tsideFactors\n", k, j, (int) LOCAL_PART(x.A), N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, "   %d\t\t%d\t%d\t%d\n", i,
	    x.base[i], x.stride[i], x.sideFactors[i]);
  }
  fprintf(stderr, "is [%x]\n  Array<%d>:\tbase\tstride\tsideFactors\n",
	  (int) LOCAL_PART(r.A), N - 1);
  for (i = 0; i < N - 1; i++) {
    fprintf(stderr, "   %d\t\t%d\t%d\t%d\n", i,
	    r.base[i], r.stride[i], r.sideFactors[i]);
  }
#endif

  return r;
}
#else
void SLICE_METHOD(const ti_ARRAY x, jint k, jint j)
{
  fputs("Error: Slice on 1d array (ignored)\n", stderr);
}
#endif

#if assignable
/* x.exchange(w) is equivalent to:
   ordered_foreach (p in Ti.myTeam().domain()) x[p] = broadcast w from p[1];

   The current implementation assumes myTeam().domain() is [0 : PROCS - 1].
   */
void EXCHANGE_METHOD(const ti_ARRAY x, T w)
{
  assertExists(x, "(in exchange method)");
  if (N != 1) fprintf(stderr, "Error: exchange() used on %dd array; "
		      "it should only be used on 1d arrays.\n", N);
#if N_IS_1
  {
    jint p1;

    for (p1 = 0; p1 < PROCS; p1++) {
      T receive;
      {
	BROADCAST_BEGIN( T, p1 );	
	if (p1 == MYPROC)
	  BROADCAST_SEND(    receive, T, w );
	else
	  BROADCAST_RECEIVE( receive, T, p1 );
	BROADCAST_END( receive, T, p1 );
      }
      {
	ti_POINT p;
	PTR_TO_T target;
	cl_ti_domains_tiPoint1_construct(p, p1);
	INDEX(target, x.A, ti_conv("(in exchange method)", &x, p));
	assign(target, receive);
      }
    }
  }
#endif
  /*
   * Ensure that every process is done writing into its respective
   * array before continuing.  Otherwise, one process can see
   * another process's array in a partially-written state.
   */
  barrier();
}

void COPY_METHOD(const ti_ARRAY x, const ti_ARRAY y)
{
  int i, lo[N], hi[N], stride[N];
  ti_POINT p, q, s;
  ti_RECTDOMAIN R;

  assertExists(x, "(in copy method)");
  assertExists(y, "(in copy method)");
  R = RECTDOMAIN_INTERSECTION(y.domain, x.domain);
  if (RECTDOMAIN_ISNULL(R)) return;
  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  
  {
    T tmp;
    forall2(e, x, f, y, lo, hi, stride, deref(tmp, f); weak_assign(e, tmp));
  }

#ifdef USE_SPLIT_C
  sync();
#endif
}


/*
 * x.set(w) is equivalent to:
 * foreach (p in x.domain()) x[p] = w;
 */
INLINE_MODIFIER void SET_METHOD(const ti_ARRAY x, T w)
{
  ti_RECTDOMAIN R = x.domain;
  ti_POINT p, q, s;
  jint lo[N], hi[N], stride[N], i;

  assertExists(x, "(in set method)");
  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  
  forall(e, x, lo, hi, stride, assign(e, w));
}
#endif /* assignable */

INLINE_MODIFIER ti_ARRAY TRANSLATE_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  int i;
  assertExists(x, "(in translate method)");
  n.domain = RECTDOMAIN_TRANSLATE(x.domain, p);
  p = RECTDOMAIN_MIN(n.domain);
  for (i = 0; i < N; i++)
    n.base[i] = POINT_INDEX(p, i + 1);

#if DEBUG_ARRAYS
#if 0
  fprintf(stderr, "p%d: translated array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(n.domain), (int) TO_LOCAL(n.domain));
#endif
  PRINT_DOMAIN(n.domain, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%d\n", i,
	    n.base[i], n.stride[i], n.sideFactors[i]);
  }
#endif

  return n;
}

#define PERMUTE_C_ARRAY(src, dest, p)						\
do {										\
  int i_;									\
										\
  for (i_ = 0; i_ < N; i_++) dest[POINT_INDEX(p, i_ + 1) - 1] = src[i_];	\
} while (0)

INLINE_MODIFIER ti_ARRAY PERMUTE_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  assertExists(x, "(in permute method)");
  n.domain = RECTDOMAIN_PERMUTE(n.domain, p);
  PERMUTE_C_ARRAY(x.base, n.base, p);
  PERMUTE_C_ARRAY(x.stride, n.stride, p);
  PERMUTE_C_ARRAY(x.sideFactors, n.sideFactors, p);
  return n;
}

#undef PERMUTE_C_ARRAY

INLINE_MODIFIER ti_ARRAY INJECT_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  int i;
  assertExists(x, "(in inject method)");
  n.domain = RECTDOMAIN_INJECT(x.domain, p);
  for (i = 0; i < N; i++) {
    int px = POINT_INDEX(p, i + 1);
    n.base[i] *= px;
    n.stride[i] *= px;
  }      

#if 0
#if 0
  fprintf(stderr, "p%d: injected array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(n.domain), (int) TO_LOCAL(n.domain));
#endif
  PRINT_DOMAIN(n.domain, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%d\n", i,
	    n.base[i], n.stride[i], n.sideFactors[i]);
  }
#endif

  return n;
}

INLINE_MODIFIER ti_ARRAY PROJECT_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  int i;
  assertExists(x, "(in project method)");
  n.domain = RECTDOMAIN_PROJECT(x.domain, p);
  for (i = 0; i < N; i++) {
    int px = POINT_INDEX(p, i + 1);
    assert(("project() beyond limits", n.base[i] % px == 0));
    n.base[i] /= px;
    assert(("project() beyond limits", n.stride[i] % px == 0));
    n.stride[i] /= px;
  }      

#if 1
#if 0
  fprintf(stderr, "p%d: injected array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(n.domain), (int) TO_LOCAL(n.domain));
#endif
  PRINT_DOMAIN(n.domain, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%d\n", i,
	    n.base[i], n.stride[i], n.sideFactors[i]);
  }
#endif

  return n;
}

#undef POINT_PRINT
#undef N_LARGER_THAN_1 N_MINUS_1
#undef LOCAL_PART
#undef PROC_PART
#undef deref
#undef INDEX
#undef assign
#undef weak_assign
#undef assignable
#undef array_ralloc
#undef array_malloc
