/* $Id: src-input.cc 1.12 Thu, 04 Mar 1999 16:34:43 -0800 liblit $ */

// src-input.cc: Input of classes and packages by reading in the source

#include <cerrno>
#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <dirent.h>
#include <fstream>
#include <iostream>
#include <string>
#include <sys/stat.h>
#include "AST.h"
#include "ClassDecl.h"
#include "NamedPath.h"
#include "UnnamedPath.h"
#include "compiler.h"
#include "decls.h"
#include "errors.h"
#include "utils.h"


extern int DEBUG;

static SearchPath &pickLib(Decl *container)
{
  static NamedPath   NamedLib;
  static UnnamedPath UnnamedLib;

  return (container == unnamedPackage)
    ? (SearchPath &) UnnamedLib
    : (SearchPath &) NamedLib;
}


void PackageDecl::readIn()
{
  bool empty = true;

  if (_container == NULL)
    _environ = new Environ (NULL);
  else
    _environ = new Environ (_container->environ());
  isStub = false;

  const SearchPath &paths = pickLib(this);
  string subdir = fullName('/');
  if (!subdir.empty()) subdir += '/';
  
  for (SearchPath::const_iterator path = paths.begin();
       path != paths.end(); ++path) {
    const string dirName = *path + subdir;
    DIR *dir = opendir (dirName.c_str());
    if (dir) {
      for (dirent *entry = readdir (dir); entry; entry = readdir (dir)) {
	const string name = entry->d_name;
	const string::size_type length = name.size();
	string className;
	  
	if (name == "." || name == "..")
	  continue;
	else if (length > 3 && name.substr(length - 3) == ".ti")
	  className = name.substr(0, length - 3);
	else if (length > 5 && name.substr(length - 5) == ".java")
	  className = name.substr(0, length - 5);
	  
	if (!className.empty()) {
	  _environ->add(new ClassDecl(intern(className), this));
	  empty = false;
	  if (DEBUG)
	    cerr << errorName() << " : found source in " << dirName << name << endl;
	    
	} else {
	  string fullname = dirName + name;
	  struct stat fs;
	      
	  if (stat(fullname.c_str(), &fs) == 0 && S_ISDIR(fs.st_mode)) {
	    _environ->add(new PackageDecl(intern(name), this));
	    empty = false;
	    if (DEBUG)
	      cerr << errorName() << " : found subpackage in " << fullname << endl;
	  }
	}
      }
      closedir (dir);
    } else if (errno != ENOENT)
      Warning ("could not read directory " + dirName + ": " + strerror(errno));
  }

  if (empty && this != unnamedPackage)      
    Warning ("unable to find any sources or subpackages for " + errorName());
}

// debugging globals to make sure that loads, environment builds happen
// at the "right" time
bool allshouldbeloaded;
bool canbuildenv;

void ClassDecl::drequire()
{
  if (!_source)
    {
      _source = TreeNode::omitted;
      assert(!allshouldbeloaded);

      string fileName = fullName('/');
      FILE * const file = pickLib(container()).openSource(fileName);

      if (DEBUG)
	cerr << ">Reading in class " << fullName() << " from " << fileName << endl;
      
      //load(fileName, file, isTitanium(fileName) || fullName().substr(0, 3) == "ti.");
      load(fileName, file, true);
      if (file) fclose(file);

      if (_source->absent())
	Error("file " + fileName + " doesn't contain class or interface " +
	      fullName());

      if (DEBUG)
	cerr << ">Done reading class " << fullName() << endl;
    }
}

void ClassDecl::buildEnviron()
{
  if (DEBUG)
    cerr << ">Building env for class " << fullName() << endl;
  assert(!allshouldbeloaded && canbuildenv);

  drequire();
  // builds environments for all recently loaded classes, including this one
  buildEnvironments(); 
  // If class didn't load, give it a dummy environment, etc
  if (!_environ)
    {
      environ(new Environ(PackageDecl::System->environ()));
      superClass(ObjectDecl);
      visits = 1;
    }
}
